"""Plugin to view all fields available for variants, annotations and samples

SelectionQueryWidget class is seen by the user and uses ColumnQueryModel class
as a model that handles records from the database.
"""
# Qt imports
from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *

# Custom imports
from .plugin import QueryPluginWidget
from cutevariant.core import sql
from cutevariant.gui.style import TYPE_COLORS
from cutevariant.gui.ficon import FIcon
from cutevariant.commons import logger

LOGGER = logger()


class ColumnQueryModel(QStandardItemModel):
    """Model to store all fields available for variants, annotations and samples"""

    def __init__(self):
        super().__init__()
        self.setColumnCount(2)
        self.query = None
        self.items = []

    def load(self):
        """Load columns into the model"""
        self.clear()
        # Store QStandardItem as a list to detect easily which one is checked
        self.items = list()
        categories = dict()

        # Fields is a dictionnary with (name,type,description,category) as keys
        # Create a category item and add fields as children
        for record in sql.get_fields(self.query.conn):

            if record["category"] != "samples":
                item = QStandardItem(record["name"])
                item.setEditable(False)
                item.setToolTip(record["description"])
                # map value type to color
                item.setIcon(FIcon(0xF70A, TYPE_COLORS[record["type"]]))
                item.setCheckable(True)
                item.setData(record)
                self.items.append(item)

                # Create category parent items
                if record["category"] not in categories.keys():
                    cat_item = QStandardItem(record["category"])
                    cat_item.setEditable(False)
                    cat_item.setIcon(FIcon(0xF645))
                    self.appendRow(cat_item)
                    categories[record["category"]] = cat_item

        # Append child to parent
        for item in self.items:
            category = item.data()["category"]
            if category != "sample":
                categories[category].appendRow(item)

        # Append samples
        sample_cat_item = QStandardItem("samples")
        sample_cat_item.setEditable(False)
        sample_cat_item.setIcon(FIcon(0xF645))
        self.appendRow(sample_cat_item)

        # Get sample names
        samples_names = (sample["name"] for sample in sql.get_samples(self.query.conn))

        for sample_name in samples_names:
            sample_item = QStandardItem(sample_name)
            sample_item.setCheckable(True)
            sample_item.setIcon(FIcon(0xF2E6))
            sample_item.setData({"name": ("genotype", sample_name, "GT")})
            sample_cat_item.appendRow(sample_item)
            self.items.append(sample_item)

    def check_query_columns(self):
        """Check column name if it is in query.columns"""
        for item in self.items:
            item.setCheckState(Qt.Unchecked)
            if item.data()["name"] in self.query.columns:
                item.setCheckState(Qt.Checked)


class ColumnQueryWidget(QueryPluginWidget):
    """Display all fields according categories"""

    def __init__(self):
        super().__init__()

        self.setWindowTitle(self.tr("Columns"))
        self.view = QTreeView()
        self.model = ColumnQueryModel()
        self.view.setModel(self.model)
        # self.view.setIndentation(0)
        self.view.header().setVisible(False)
        layout = QVBoxLayout()

        layout.addWidget(self.view)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)
        self.model.itemChanged.connect(self.on_item_checked)

    def on_init_query(self):
        """Overrided"""
        self.model.query = self.query
        self.model.load()

    def on_change_query(self):
        """Overrided"""
        #  Avoid crash with infinite loop by disconnecting the signals
        self.model.itemChanged.disconnect(self.on_item_checked)

        # check selected query fields
        self.model.check_query_columns()

        # Reconnect signals
        self.model.itemChanged.connect(self.on_item_checked)

    def on_item_checked(self):
        """Called when an item has been checked"""

        # Get selected columns from checked items
        selected_columns = [
            item.data()["name"]
            for item in self.model.items
            if item.checkState() == Qt.Checked
        ]

        # Update query with selected columns
        self.query.columns = selected_columns

        LOGGER.debug(
            "ColumnQueryWidget::on_item_checked: columns: %s", self.query.columns
        )

        # Signal other widget that query has changed
        self.query_changed.emit()



"""Widget to display variants from a query in the GUI.

ViewQueryWidget class is added to a QTabWidget and is seen by the user;
it uses QueryModel class as a model that handles records from the database.
QueryDelegate class handles the aesthetic of the view.
"""

# Standard imports
import copy
import csv

# Qt imports
from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *

# Custom imports
from cutevariant.gui.ficon import FIcon
from .plugin import QueryPluginWidget
from cutevariant.core import Query
from cutevariant.core import sql
from cutevariant.gui import style
from cutevariant.gui.infovariantwidget import VariantPopupMenu
from cutevariant.commons import logger
from cutevariant.commons import GENOTYPE_ICONS

LOGGER = logger()


class QueryModel(QAbstractItemModel):
    """
    QueryModel is a Qt model class which contains variants datas from sql Query. 
    It loads paginated data from Query and create an interface for a Qt view.
    When a variant belong to many transcripts, duplicate variants are grouped by (chr,pos,ref,alt) 
    and stored as a Tree structure. Variant row can then be expanded to display each transcript annotation. 

    See Qt model/view programming for more information
    https://doc.qt.io/qt-5/model-view-programming.html

    Variants are stored internally as a list of variants. By default, there is only one transcript per row. 
    When user expand the row, it will append duplicates variants as children. 
    For example, this is a tree with 2 variants , each of them refer to many transcripts. 

    self.variants = [ 
        [(chr1,2434,A,T, transcriptA),(chr1,2434,A,T, transcriptB),(chr1,2434,A,T, transcriptC)],
        [(chr1,9999,C,T, transcriptA),(chr1,9999,C,T, transcriptB),(chr1,9999,C,T, transcriptC),(chr1,9999,C,T, transcriptD]
        ]
    ]

    ├── chr1,2434,A,T, transcriptA  # Cannonical transcripts
    │   ├── chr1,2434,A,T, transcriptB
    │   ├── chr1,2434,A,T, transcriptC
    ├── chr1,9999,C,T, transcriptA # Cannonical transcripts
    │   ├── chr1,9999,C,T, transcriptB
    │   ├── chr1,9999,C,T, transcriptC
    │   ├── chr1,9999,C,T, transcriptD

    Attributes:
        limit(int): Variant count to display. See SQL LIMIT
        page(int): current page to display. See SQL OFFSET
        total(int): Total variant count
        variants(list): Internal data to store variants

    Example:
        model = QueryModel()
        view = QTreeView()
        view.setModel(model)
        model.setQuery(query)

    """

    def __init__(self, parent=None):
        super().__init__()
        self.limit = 50
        self.page = 0
        self.total = 0
        self._query = None
        self.variants = []

    def level(self, index: QModelIndex) -> bool:
        """Return level of index. 

        Args:
            index (QModelIndex): The model index

        Returns: 
            int: 0 if index is the invisible root , 1 if index is a variant parent, 2 if index is a variant child  
        """

        if index == QModelIndex():
            return 0

        if index.parent() == QModelIndex():
            return 1

        if index.parent().parent() == QModelIndex():
            return 2

    def rowCount(self, parent=QModelIndex()):
        """Overrided : Return children count of index 
        """
        #  If parent is root
        if self.level(parent) == 0:
            return len(self.variants)

        if self.level(parent) == 1:
            return len(self.variants[parent.row()]) - 1  # Omit first

    def columnCount(self, parent=QModelIndex()):
        """Overrided: Return column count of parent . 

        Parent is not used here. 
        """

        # If no query is defined, return nothing
        if not self._query:
            return 0

        #  return query columns + child count columns
        #  children count - chr - pos .....
        return len(self._query.columns) + 1

    def index(self, row: int, column: int, parent=QModelIndex()) -> QModelIndex:
        """Overrided: Create a new model index from row, column and parent  
        
        Args:
            row (int): row number
            column (int): column number
            parent (QModelIndex): parent index

        """

        # avoid error
        if not self.hasIndex(row, column, parent):
            return QModelIndex()

        # Create index for variant as parent
        if self.level(parent) == 0:
            # createIndex take None as internalId. @see self.parent()
            return self.createIndex(row, column, None)

        # Create index for variant as child
        if self.level(parent) == 1:
            # createIndex take parent.row() as internalId . @see self.parent()
            return self.createIndex(row, column, parent.row())

    def parent(self, index: QModelIndex()) -> QModelIndex:
        """Overrided: Return the parent of index """

        #  avoid error
        if not index.isValid():
            return QModelIndex()

        #  If internalId is None, index is a variant parent
        if index.internalId() == None:
            return QModelIndex()

        # Otherwise, index is a variant child at position row=internalid in the parent
        else:
            return self.index(index.internalId(), 0, QModelIndex())

    def data(self, index: QModelIndex(), role=Qt.DisplayRole):
        """ Overrided: return index data according role.
        This method is called by the Qt view to get data to display according Qt role. 
        
        Params:
            index (QModelIndex): index from where your want to get data 
            role (Qt.ItemDataRole): https://doc.qt.io/qt-5/qt.html#ItemDataRole-enum

        Examples:
            index = model.index(row=10, column = 1)
            chromosome_value = model.data(index)
        """

        # Avoid error
        if not index.isValid():
            return None

        #  ---- Display Role ----
        if role == Qt.DisplayRole:
            # Return data for the first level
            if self.level(index) == 1:
                # First column correspond to children count
                if index.column() == 0:
                    return str(self.children_count(index))
                # Other data come from internal self.variants list
                else:
                    return str(self.variant(index)[index.column()])
                    # return str(self.variants[index.row()][0][index.column()])

            # Return data for the second level
            if self.level(index) == 2:
                if index.column() == 0:
                    # No children for the first columns
                    return ""
                else:
                    #  Display children data
                    return str(self.variant(index)[index.column()])
        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Overrided: Return column name 
        This method is called by the Qt view to display vertical or horizontal header data.

        Params:
            section (int): row or column number depending on orientation
            orientation (Qt.Orientation): Qt.Vertical or Qt.Horizontal
            role (Qt.ItemDataRole): https://doc.qt.io/qt-5/qt.html#ItemDataRole-enum

        Examples:
            # return 4th column name 
            column_name = model.headerData(4, Qt.Horizontal)

         """

        #  Display columns headers
        if orientation == Qt.Horizontal:
            if role == Qt.DisplayRole:
                if section == 0:
                    return "children"
                else:
                    col = self._query.columns[section - 1]
                    if type(col) == tuple and len(col) == 3:  #  show functions
                        fct, arg, field = col
                        return f"{fct}({arg}).{field}"
                    else:
                        return self._query.columns[section - 1]
        return None

    def hasChildren(self, parent: QModelIndex) -> bool:
        """Overrided: Return True if parent index has children """

        # if invisible root node, always return True
        if parent == QModelIndex():
            return True

        if self.level(parent) == 1:
            children_count = self.children_count(parent)
            return children_count >= 1

        return False

        # if parent.parent() == QModelIndex():
        #     return self._children_count(parent) > 1

    def canFetchMore(self, parent: QModelIndex) -> bool:
        """ Overrided : Return True if children fetching is required """
        return self.hasChildren(parent)

    def fetchMore(self, parent: QModelIndex):
        """Overrided: Fetch children variant 
        """

        # avoid error
        if parent == QModelIndex():
            return

        #  get sql variant id
        variant_id = self.variants[parent.row()][0][0]

        columns = ",".join(self._query.get_columns(do_not_add_default_things=True))
        joints = self._query.get_joints()

        #  TODO : need to put this into QUERY
        # TODO: add filter into left join annotations ... instead in where
        sub_query = f"""SELECT variants.id, {columns} FROM variants
        {joints}
        WHERE annotations.variant_id = {variant_id}"""
        LOGGER.debug(
            "QueryModel:fetchMore:: Extra children cols sub query %s", sub_query
        )

        records = list(self._query.conn.cursor().execute(sub_query).fetchall())

        # remove last item because it's the same as parent
        records.pop()

        # Insert children
        self.beginInsertRows(parent, 0, len(records))

        # Clear pevious children
        self.variants[parent.row()][1:] = []

        for idx, record in enumerate(records):  # skip first records
            self.variants[parent.row()].append(tuple(record))

        self.endInsertRows()

    def children_count(self, index: QModelIndex) -> int:
        """Return children count from variant

        This one is the last value of sql record output and correspond to the COUNT(annotation)
        of the GROUP BY
        """
        return self.variants[index.row()][0][-1] - 1

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, query: Query):
        # PS: default group by method: ("chr","pos","ref","alt")
        self._query = query

    def load(self):
        """Load variant data into the model from query attributes

        Called by:
            - on_change_query() from the view.
            - sort() and setPage() by the model.
        """
        self.beginResetModel()
        # Set total of variants for pagination
        self.total = self.query.variants_count()

        # Append a list because child can be append after
        self.variants = [
            [tuple(variant)]
            for variant in self._query.items(self.limit, self.page * self.limit)
        ]

        LOGGER.debug("QueryModel:load:: variants queried\n%s", self.variants)
        self.endResetModel()

    def hasPage(self, page: int) -> bool:
        """ Return True if <page> exists otherwise return False """
        return page >= 0 and page * self.limit < self.total

    def setPage(self, page: int):
        """ set the page of the model """
        if self.hasPage(page):
            self.page = page
            self.load()

    def nextPage(self):
        """ Set model to the next page """
        if self.hasPage(self.page + 1):
            self.setPage(self.page + 1)

    def previousPage(self):
        """ Set model to the previous page """
        if self.hasPage(self.page - 1):
            self.setPage(self.page - 1)

    def firstPage(self):
        """ Set model to the first page """
        self.setPage(0)

    def lastPage(self):
        """ Set model to the last page """
        self.setPage(int(self.total / self.limit))

    def sort(self, column: int, order):
        """Overrided: Sort data by specified column 
        
        column (int): column id 
        order (Qt.SortOrder): Qt.AscendingOrder or Qt.DescendingOrder 

        """
        if column < self.columnCount():
            colname = self._query.columns[column - 1]

            self._query.order_by = colname
            self._query.order_desc = order == Qt.DescendingOrder
            self.load()

    def displayed(self):
        """Get ids of first, last displayed variants on the total number

        :return: Tuple with (first_id, last_id, self.total).
        :rtype: <tuple <int>,<int>,<int>>
        """
        first_id = self.limit * self.page

        if self.hasPage(self.page + 1):
            # Remainder : self.total - (self.limit * (self.page + 1)))
            last_id = self.limit * (self.page + 1)
        else:
            # Remainder : self.total - (self.limit * self.page)
            last_id = self.total

        return (first_id, last_id, self.total)

    def variant(self, index: QModelIndex):
        """ Return variant data according index 
        
        Examples:
            variant = model.variant(index)
            print(variant) # ["chr","242","A","T",.....]

        """

        if self.level(index) == 1:
            return self.variants[index.row()][0]

        if self.level(index) == 2:
            return self.variants[index.parent().row()][
                index.row() + 1
            ]  #  + 1 because the first element is the parent


class QueryDelegate(QStyledItemDelegate):
    """
    This class specify the aesthetic of the view

    styles and color of each variant displayed in the view are setup here


    """

    def background_color_index(self, index):
        """ return background color of index """

        base_brush = qApp.palette("QTreeView").brush(QPalette.Base)
        alternate_brush = qApp.palette("QTreeView").brush(QPalette.AlternateBase)

        if index.parent() == QModelIndex():
            if index.row() % 2:
                return base_brush
            else:
                return alternate_brush

        if index.parent().parent() == QModelIndex():
            return self.background_color_index(index.parent())

        return base_brush

    def paint(self, painter, option, index):
        """
        overriden

        Draw the view item for index using a painter

        """

        palette = qApp.palette("QTreeView")
        #  get column name of the index
        colname = index.model().headerData(index.column(), Qt.Horizontal)

        #  get value of the index
        value = index.data(Qt.DisplayRole)

        # get select sate
        select = option.state & QStyle.State_Selected

        #  draw selection if it is
        if not select:
            bg_brush = self.background_color_index(index)
        else:
            bg_brush = palette.brush(QPalette.Highlight)

        painter.save()
        painter.setBrush(bg_brush)
        painter.setPen(Qt.NoPen)
        painter.drawRect(option.rect)
        painter.restore()

        # painter.setPen(pen)

        alignement = Qt.AlignLeft | Qt.AlignVCenter

        # # Add margin for first columns if index is first level
        # if index.column() == 0 and index.parent() == QModelIndex():

        #     expanded = bool(option.state & QStyle.State_Open)

        #     branch_option = copy.copy(option)
        #     branch_option.rect.setWidth(65)

        #     qApp.style().drawPrimitive(QStyle.PE_IndicatorBranch, branch_option, painter)

        #     icon = index.data(Qt.DecorationRole)
        #     if icon:
        #         target = QRect(0,0, option.decorationSize.width(), option.decorationSize.height())
        #         target.moveCenter(option.rect.center())
        #         painter.drawPixmap(option.rect.x()+5, target.top() ,icon.pixmap(option.decorationSize))

        # if index.column() == 0:
        #     option.rect.adjust(40,0,0,0)

        # Draw cell depending column name
        if colname == "impact":
            painter.setPen(
                QPen(style.IMPACT_COLOR.get(value, palette.color(QPalette.Text)))
            )
            painter.drawText(option.rect, alignement, str(index.data()))
            return

        if colname == "gene":
            painter.setPen(QPen(style.GENE_COLOR))
            painter.drawText(option.rect, alignement, str(index.data()))
            return

        if "genotype" in colname:
            val = int(value)

            icon_code = GENOTYPE_ICONS.get(val, -1)
            icon = FIcon(icon_code, Qt.white).pixmap(20, 20)
            painter.setRenderHint(QPainter.Antialiasing)
            painter.drawPixmap(option.rect.left(), option.rect.center().y() - 8, icon)
            return

        if "consequence" in colname:
            painter.save()
            painter.setClipRect(option.rect, Qt.IntersectClip)
            painter.setRenderHint(QPainter.Antialiasing)
            soTerms = value.split("&")
            rect = QRect()
            font = painter.font()
            font.setPixelSize(10)
            painter.setFont(font)
            metrics = QFontMetrics(painter.font())
            rect.setX(option.rect.x())
            rect.setY(option.rect.center().y() - 5)

            #  Set background color according so terms
            #  Can be improve ... Just a copy past from c++ code
            bg = "#6D7981"
            for so in soTerms:
                for i in style.SO_COLOR.keys():
                    if i in so:
                        bg = style.SO_COLOR[i]

                painter.setPen(Qt.white)
                painter.setBrush(QBrush(QColor(bg)))
                rect.setWidth(metrics.width(so) + 8)
                rect.setHeight(metrics.height() + 4)
                painter.drawRoundedRect(rect, 3, 3)
                painter.drawText(rect, Qt.AlignCenter, so)

                rect.translate(rect.width() + 4, 0)

            painter.restore()
            return

        painter.setPen(
            QPen(palette.color(QPalette.HighlightedText if select else QPalette.Text))
        )
        painter.drawText(option.rect, alignement, str(index.data()))

    def draw_biotype(self, value):
        pass

    def sizeHint(self, option, index):
        """Override: Return row height"""
        return QSize(0, 30)


class QueryTreeView(QTreeView):
    def __init__(self, parent=None):
        super().__init__()

    def drawBranches(self, painter, rect, index):
        """ overrided : Draw Branch decorator with background 
        
        Backround is not alternative for children but inherits from parent 

        """
        painter.save()
        painter.setPen(Qt.NoPen)
        painter.setBrush(self.itemDelegate().background_color_index(index))
        painter.drawRect(rect)

        if index.parent() != QModelIndex():
            #  draw child indicator
            painter.drawPixmap(rect.center(), FIcon(0xF12F).pixmap(10, 10))

        painter.restore()

        super().drawBranches(painter, rect, index)


class ViewQueryWidget(QueryPluginWidget):
    """Contains the view of query with several controller"""

    variant_clicked = Signal(dict)

    def __init__(self):
        super().__init__()
        self.model = QueryModel()
        self.delegate = QueryDelegate()
        self.setWindowTitle(self.tr("Variants"))
        self.topbar = QToolBar()
        self.bottombar = QToolBar()
        self.view = QueryTreeView()

        # self.view.setFrameStyle(QFrame.NoFrame)
        self.view.setModel(self.model)
        self.view.setItemDelegate(self.delegate)
        # self.view.setAlternatingRowColors(True)
        self.view.setSortingEnabled(True)
        self.view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.view.setSelectionMode(QAbstractItemView.ContiguousSelection)
        self.view.setRootIsDecorated(True)  # Manage from delegate
        # self.view.setIndentation(0)
        self.view.setIconSize(QSize(22, 22))
        self.view.setAnimated(True)
        self.view.setStyleSheet("QAbstractScrollArea {border-left: 20px solid red}")

        # self.view.setItemDelegate(self.delegate)

        main_layout = QVBoxLayout()
        main_layout.addWidget(self.topbar)
        main_layout.addWidget(self.view)
        main_layout.addWidget(self.bottombar)
        main_layout.setContentsMargins(0, 0, 0, 0)

        # Construct top bar
        # These actions should be disabled until a query is made (see query setter)
        self.export_csv_action = self.topbar.addAction(
            self.tr("Export variants"), self.export_csv
        )
        self.export_csv_action.setEnabled(False)

        # Add spacer to push next buttons to the right
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.topbar.addWidget(spacer)

        # Add combobox to choose the grouping method of variants

        # Construct bottom bar
        # These actions should be disabled until a query is made (see query setter)
        self.page_info = QLabel()
        self.page_box = QLineEdit()
        self.page_box.setReadOnly(False)
        self.page_box.setValidator(QIntValidator())
        # self.page_box.setFrame(QFrame.NoFrame)
        self.page_box.setFixedWidth(50)
        self.page_box.setAlignment(Qt.AlignHCenter)
        self.page_box.setStyleSheet("QWidget{background-color: transparent;}")
        self.page_box.setText("0")
        self.page_box.setFrame(QFrame.NoFrame)
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        # Setup actions
        self.show_sql_action = self.bottombar.addAction(
            FIcon(0xF865), self.tr("See SQL query"), self.show_sql
        )
        self.show_sql_action.setEnabled(False)
        self.bottombar.addWidget(self.page_info)
        self.bottombar.addWidget(spacer)
        self.bottombar.addAction(FIcon(0xF792), "<<", self.model.firstPage)
        self.bottombar.addAction(FIcon(0xF04D), "<", self.model.previousPage)
        self.bottombar.addWidget(self.page_box)
        self.bottombar.addAction(FIcon(0xF054), ">", self.model.nextPage)
        self.bottombar.addAction(FIcon(0xF793), ">>", self.model.lastPage)

        self.bottombar.setIconSize(QSize(16, 16))
        self.bottombar.setMaximumHeight(30)

        self.bottombar.setContentsMargins(0, 0, 0, 0)

        self.setLayout(main_layout)

        self.model.modelReset.connect(self.updateInfo)

        # Create menu
        self.context_menu = VariantPopupMenu()

        # emit variant when clicked
        self.view.clicked.connect(self._variant_clicked)
        self.page_box.returnPressed.connect(self._update_page)

    def on_init_query(self):
        """Method override from AbstractQueryWidget"""
        self.export_csv_action.setEnabled(True)
        self.show_sql_action.setEnabled(True)
        self.model.query = self.query

    def on_change_query(self):
        """Method override from AbstractQueryWidget"""
        #  reset current page
        self.model.page = 0
        self.model.load()
        self.view.header().setSectionResizeMode(0, QHeaderView.ResizeToContents)

    def updateInfo(self):
        """Update metrics for the current query

        .. note:: Update page_info and page_box.
        """

        # Set text
        self.page_info.setText(
            self.tr("{} variant(s)  {}-{} of {}").format(
                self.model.total, *self.model.displayed()
            )
        )
        page_box_text = str(self.model.page)
        self.page_box.setText(page_box_text)

        # Adjust page_èbox size to content
        fm = self.page_box.fontMetrics()
        # self.page_box.setFixedWidth(fm.boundingRect(page_box_text).width() + 5)

    def _variant_clicked(self, index):
        """Slot called when the view (QTreeView) is clicked

        .. note:: Emit variant through variant_clicked signal.
            This signal updates InfoVariantWidget.
        .. note:: Is also called manually by contextMenuEvent() in order to
            get the variant and refresh InfoVariantWidget when the
            ContextMenuEvent is triggered.
        :return: The variant.
        :rtype: <dict>
        """
        # Get the rowid of the element at the given index
        rowid = self.model.variant(index)[0]
        # Get data from database
        variant = sql.get_one_variant(self.model.query.conn, rowid)
        # Emit variant through variant_clicked signal
        self.variant_clicked.emit(variant)
        return variant

    def export_csv(self):
        """Export variants displayed in the current view to a CSV file"""
        filepath, filter = QFileDialog.getSaveFileName(
            self,
            self.tr("Export variants of the current view"),
            "view.csv",
            self.tr("CSV (Comma-separated values) (*.csv)"),
        )

        if not filepath:
            return

        with open(filepath, "w") as f_d:
            writer = csv.writer(f_d, delimiter=",")
            # Write headers (columns in the query) + variants from the model
            writer.writerow(self.model.query.columns)
            # Duplicate the current query, but remove automatically added columns
            # and remove group by/order by commands.
            # Columns are kept as they are selected in the GUI
            query = copy.copy(self.model.query)
            query.group_by = None
            query.order_by = None
            # Query the database
            writer.writerows(
                query.conn.execute(query.sql(do_not_add_default_things=True))
            )

    def on_group_by_changed(self, index):
        """Slot called when the currentIndex in the combobox changes
        either through user interaction or programmatically

        It triggers a reload of the model and a change of the group by
        command of the query.
        """
        self.model.group_by_changed(self.group_by_combobox.currentData())

    def show_sql(self):
        box = QMessageBox()
        try:
            text = self.model.query.sql()
        except AttributeError:
            text = self.tr("No query to show")

        box.setInformativeText(text)
        box.exec_()

    def contextMenuEvent(self, event: QContextMenuEvent):
        """Overrided method: Show custom context menu associated to the current variant"""
        # Get the variant (and refresh InfoVariantWidget)
        current_index = self.view.currentIndex()
        variant = self._variant_clicked(current_index)
        # Show the context menu with the given variant
        self.context_menu.popup(variant, event.globalPos())

    def _update_page(self):
        """Set page from page_box edit. When user set a page manually, this method is called"""
        self.model.setPage(int(self.page_box.text()))

        self.page_box.clearFocus()



# Standard imports
import re

# Custom imports
from cutevariant.commons import logger

LOGGER = logger()

# Dicts of default fields depending to the type of files
# Keys are based on fields in files (in lower case),
# values are the full description of the field; the name is remaped here as
# it wil be shown in the GUI.
# PS: Consequences/annotation_impac/impacts are standardized here:
# https://www.ensembl.org/info/genome/variation/prediction/predicted_data.html#consequences
SNPEFF_ANNOTATION_DEFAULT_FIELDS = {
    "annotation": {
        "name": "consequence",
        "category": "annotations",
        "description": "consequence type",
        "type": "str",
    },
    "annotation_impact": {
        "name": "impact",
        "category": "annotations",
        "description": "impact rating of variant",
        "type": "str",
    },
    "gene_name": {
        "name": "gene",
        "category": "annotations",
        "description": "gene name",
        "type": "str",
    },
    "gene_id": {
        "name": "gene_id",
        "category": "annotations",
        "description": "Ensembl stable ID of affected gene",
        "type": "str",
    },
    "feature_id": {
        "name": "transcript",
        "category": "annotations",
        "description": "transcript name",
        "type": "str",
    },
    "transcript_biotype": {
        "name": "biotype",
        "category": "annotations",
        "description": " biotype",
        "type": "str",
    },
    "hgvs.p": {
        "name": "hgvs_p",
        "category": "annotations",
        "description": "protein hgvs",
        "type": "str",
    },
    "hgvs.c": {
        "name": "hgvs_c",
        "category": "annotations",
        "description": "coding hgvs",
        "type": "str",
    },
    "cdna.pos / cdna.length": {
        "name": "cdna_pos",
        "category": "annotations",
        "description": "relative position of base pair in cDNA sequence",
        "type": "str",
    },
    "cds.pos / cds.length": {
        "name": "cds_pos",
        "category": "annotations",
        "description": "relative position of base pair in coding sequence",
        "type": "str",
    },
    "aa.pos / aa.length": {
        "name": "aa_pos",
        "category": "annotations",
        "description": "amino acid pos",
        "type": "str",
    },
    "errors / warnings / info": {
        "name": "log",
        "category": "annotations",
        "description": "amino acid pos",
        "type": "str",
    },
}


VEP_ANNOTATION_DEFAULT_FIELDS = {
    "allele": {
        "name": "allele",
        "category": "annotations",
        "description": "the variant allele used to calculate the consequence",
        "type": "str",
    },
    "consequence": {
        "name": "consequence",
        "category": "annotations",
        "description": "consequence type",
        "type": "str",
    },
    "impact": {
        "name": "impact",
        "category": "annotations",
        "description": "impact rating of the variant",
        "type": "str",
    },
    "symbol": {
        "name": "gene",
        "category": "annotations",
        "description": "gene name",
        "type": "str",
    },
    "gene": {
        "name": "gene_id",
        "category": "annotations",
        "description": "Ensembl stable ID of affected gene",
        "type": "str",
    },
    "feature": {
        "name": "transcript",
        "category": "annotations",
        "description": "Ensembl stable ID of feature",
        "type": "str",
    },
    "feature_type": {
        "name": "feature_type",
        "category": "annotations",
        "description": "type of feature. Currently one of Transcript, RegulatoryFeature, MotifFeature.",
        "type": "str",
    },
    "biotype": {
        "name": "biotype",
        "category": "annotations",
        "description": " biotype",
        "type": "str",
    },
    "hgvsp": {
        "name": "hgvs_p",
        "category": "annotations",
        "description": "protein hgvs",
        "type": "str",
    },
    "hgvsc": {
        "name": "hgvs_c",
        "category": "annotations",
        "description": "coding hgvs",
        "type": "str",
    },
    "cdna_position": {
        "name": "cdna_pos",
        "category": "annotations",
        "description": "relative position of base pair in cDNA sequence",
        "type": "str",
    },
    "cds_position": {
        "name": "cds_pos",
        "category": "annotations",
        "description": "relative position of base pair in coding sequence",
        "type": "str",
    },
    "protein_position": {
        "name": "aa_pos",
        "category": "annotations",
        "description": "relative position of amino acid in protein",
        "type": "str",
    },
    "amino_acids": {
        "name": "amino_acids",
        "category": "annotations",
        "description": "Reference and variant amino acids; only given if the variant affects the protein-coding sequence",
        "type": "str",
    },
    "codons": {
        "name": "codons",
        "category": "annotations",
        "description": "Reference and variant codon sequence",
        "type": "str",
    },
    "existing_variation": {
        "name": "existing_variation",
        "category": "annotations",
        "description": "Identifier(s) of co-located known variants",
        "type": "str",
    },
}


class BaseParser:
    """Base class that brings together common functions of VepParser and SnpEffParser
    """

    def __init__(self):

        # Receive a dict to map default field names:
        # external to internal with descriptions
        # { external_name: { name: internal_name, category: ..., description: ...
        self.annotation_default_fields = dict()

        # Help to remove duplicated fields in annotations
        # If an annotation field is found in this set by the annotation parser
        # BaseParser.handle_descriptions(), it will not be yielded and thus not
        # used by the program.
        self.variant_field_names = set()

    def handle_descriptions(self, raw_fields: list):
        """Construct annotation_field_name with the fields of the file, and
        yield fields (dictionnaries) with the full description of fields of the file.

        :Example:
            If 'protein_position' field is encountered in a VEP file,
            'self.annotation_field_name' attribute will contain:
            ['protein_position',] and the followwing dictionnary will be yielded:
            {
                "name": "aa_pos",
                "category": "annotations",
                "description": "amino acid pos",
                "type": "str",
            }

            If a field is not provided, a default dictionary with less
            information is returned:
            {
                "name": <field_name>,
                "description": "None",
                "type":"str",
                "category":"annotations"
            }

        :param raw_fields: List of fields names.
        :type raw_fields: <list>
        :rtype: <generator <dict>>
        """
        for raw_field_name in raw_fields:
            raw_field_name = raw_field_name.strip().lower()

            # Remap field name if it is in default ones
            if raw_field_name in self.annotation_default_fields:
                _f = self.annotation_default_fields[raw_field_name]
            else:
                _f = {
                    "name": raw_field_name,
                    "description": "None",
                    "type": "str",
                    "category": "annotations",
                }

            if _f["name"] in self.variant_field_names:
                # This field is already in variants fields
                # => do not use it!
                # Append None in place of the name of the field
                # with the aim to not break handle_annotations() when it splits
                # the annotation field.
                self.annotation_field_name.append(None)
                LOGGER.info(
                    "handle_descriptions: '%s' field also found in variants; skipped",
                    _f["name"],
                )
                continue

            # Append the name of the field
            self.annotation_field_name.append(_f["name"])
            # Yield full field
            yield _f

    def handle_annotations(self, annotation_key_name, variant):
        """Remove the given key from the variant dict, add "annotations" key with
        the list of annotations into the variant dict.

        .. note:: The given variant is modified in place.
        .. note:: Structure of "annotations" value:
            [{
                'annotation_field_name1': 'data1',
                'annotation_field_name2': 'data2',
                ...
            },]
            "annotations" is a list since there may be multiple annotations for
            a variant.
        """
        raw = variant.pop(annotation_key_name)

        annotations = list()
        for transcripts in raw.split(","):
            transcript = transcripts.split("|")

            if len(self.annotation_field_name) != len(transcript):
                LOGGER.error(
                    "BaseParser:handle_annotations:: Missing field in the "
                    "annotations of the following variant:\n%s\n"
                    "These annotations will be skipped!",
                    variant,
                )
                continue

            annotation = {
                field_name: transcript[idx]
                for idx, field_name in enumerate(self.annotation_field_name)
                # Remove duplicated fields in variants, see handle_descriptions()
                if field_name is not None
            }
            annotations.append(annotation)

        # Avoid setting empty list to the variant => generates a SQL query issue
        if annotations:
            variant["annotations"] = annotations


class VepParser(BaseParser):
    """Parser of VEP annotations

    .. note:: We assume that the description field looks like this:
        ##INFO=<ID=CSQ,Number=.,Type=String,Description="Consequence annotations from Ensembl VEP. Format: Allele|Consequence|IMPACT|SYMBOL|Gene|Feature_type|Featur.."
    """

    def __init__(self):
        super().__init__()
        # Dict of dicts
        # annotation field name as keys, descriptions (name/value) as values
        self.annotation_default_fields = VEP_ANNOTATION_DEFAULT_FIELDS

    def parse_fields(self, fields):
        """Generate fields description

        Called by a reader when its get_fields() method is called and when
        an annotation parser is set.

        This function parses special annotation field "csq"/"CSQ",
        other fields are yielded without being affected.

        .. seealso:: handle_descriptions()

        :param fields: Generator of fields.
        :type fields: <generator <dict>>
        :return: Generator of fields descriptions.
        :rtype: <generator <dict>>
        """
        self.annotation_field_name = list()
        fields = tuple(fields)
        # Help to remove duplicated fields from annotations
        self.variant_field_names = {
            field["name"] for field in fields if field["name"] != "csq"
        }

        for field in fields:
            if field["name"] == "csq":
                # Handle description field and parse annotations in it
                description = field["description"]
                # Assume description looks like this :
                # ##INFO=<ID=CSQ,Number=.,Type=String,Description="Consequence annotations from Ensembl VEP. Format: field1|field2|..."

                raw_fields = re.search("Format: (.+)", description)[1].split("|")

                # yield full remaped field
                yield from self.handle_descriptions(raw_fields)
            else:
                yield field

    def parse_variants(self, variants):
        """Generate variants data

        This function removes the key "csq" from the variants,
        and add "annotations" key with the list of annotations.

        .. seelalso:: handle_annotations()

        :param variants: Generator of variants.
        :type variants: <generator <dict>>
        :return: Generator of full variants with "annotations" key.
        :rtype: <generator <dict>>
        """
        if not hasattr(self, "annotation_field_name"):
            raise Exception("Cannot parse variant without parsing first fields")

        for variant in variants:
            if "csq" in variant:
                self.handle_annotations("csq", variant)
            yield variant


class SnpEffParser(BaseParser):
    """Parser for SnpEFF annotations

    .. note:: We assume that the description field looks like this:
        INFO=<ID=ANN,Number=.,Type=String,Description="Functional annotations: 'Allele | Annotation | Annotation_Impact | Gene_Name | Gene_ID | Feature_Type | Feature_ID | Transcript_BioType | Rank | HGVS.c | HGVS.p | cDNA.pos / cDNA.length |CDS.pos / CDS.length | AA.pos / AA.length | Distance | ERRORS / WARNINGS / INFO' ">
    """

    def __init__(self):
        super().__init__()
        # Dict of dicts
        # annotation field name as keys, descriptions (name/value) as values
        self.annotation_default_fields = SNPEFF_ANNOTATION_DEFAULT_FIELDS

    def parse_fields(self, fields):
        """Generate fields description

        Called by a reader when its get_fields() method is called and when
        an annotation parser is set.

        This function parses special annotation field "ann"/"ANN",
        other fields are yielded without being affected.

        .. seealso:: handle_descriptions()

        :Input example:
            ({
            'name': 'generic_field1',
            'description': ...
            'category': ...
            'type': ...
            },
            {
            'name': 'ann',
            'description': '... annotation_field_name1 | annotation_field_name2 ...',
            ...
            })

        :Output example:
            ({
            'name': 'generic_field1',
            'description': ...
            'category': ...
            'type': ...
            },
            {
            'name': 'annotation_field_name1',
            'description': ...
            'category': ...
            'type': ...
            },
            {
            'name': 'annotation_field_name2',
            'description': ...
            'category': ...
            'type': ...
            })

        .. note:: Names of fields are changed to lowercase.

        :param fields: Generator of fields.
        :type fields: <generator <dict>>
        :return: Generator of full fields descriptions.
        :rtype: <generator <dict>>
        """
        self.annotation_field_name = list()
        fields = tuple(fields)
        # Help to remove duplicated fields from annotations
        self.variant_field_names = {
            field["name"] for field in fields if field["name"] != "ann"
        }

        for field in fields:
            if field["name"] == "ann":
                # Handle description field and parse annotations in it
                description = field["description"]
                # Assume description looks like this :
                # INFO=<ID=ANN,Number=.,Type=String,Description="Functional annotations: 'field1 | field2 | ...' ">

                raw_fields = re.search("'(.+)'", description)[1].split("|")

                # yield full remaped field
                yield from self.handle_descriptions(raw_fields)
            else:
                # Field is not an annotation: do nothing
                yield field

    def parse_variants(self, variants):
        """Generate variants data

        This function removes the key "ann" from the variants,
        and add "annotations" key with the list of annotations.

        .. seelalso:: handle_annotations()

        :param variants: Generator of variants.
        :type variants: <generator <dict>>
        :return: Generator of full variants with "annotations" key.
        :rtype: <generator <dict>>
        """
        if not hasattr(self, "annotation_field_name"):
            raise Exception("Cannot parse variant without parsing first fields")

        for variant in variants:
            if "ann" in variant:
                # Modify the current variant:
                # remove "ann" data, replace it with "annotations"
                self.handle_annotations("ann", variant)
            yield variant



"""Expose of high-level reader classes"""
from .vcfreader import VcfReader
from .csvreader import CsvReader
from .fakereader import FakeReader
from .abstractreader import check_field_schema, check_variant_schema



from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *
import sqlite3


from .plugin import QueryPluginWidget
from cutevariant.core import sql, Query
from cutevariant import commons as cm

LOGGER = cm.logger()


class Node(object):
    def __init__(self, node_id, parent=None):
        self.parent = parent
        self.childs = []
        self.set_node_id(node_id)

    def add_child(self, child):
        child.parent = self
        self.childs.append(child)

    def child(self, row: int):
        return self.childs[row]

    def child_count(self):
        return len(self.childs)

    def fetch_count(self):
        return abs(self.right - self.left) / 2

    def row(self):
        if self.parent:
            return self.parent.childs.index(self)
        return 0

    def __repr__(self):
        return self.name + " " + self.hpo

    def set_node_id(self, node_id=None):
        self.node_id = node_id

        #  Node vide ( QModelIndex root )
        if not node_id:
            return

        req = Node.conn.execute(
            f"SELECT nodes.id as 'node_id', terms.id as 'term_id', nodes.left, nodes.right, nodes.depth, terms.name, terms.hpo FROM nodes, terms WHERE nodes.term_id = terms.id AND nodes.id = {node_id}"
        )
        result = req.fetchone()

        self.node_id = result[0]
        self.term_id = result[1]
        self.left = result[2]
        self.right = result[3]
        self.depth = result[4]
        self.name = result[5]
        self.hpo = result[6]

    def load_child(self):
        self.childs.clear()
        cursor = Node.conn.cursor()
        sql = f"SELECT nodes.id FROM nodes WHERE nodes.left > {self.left} AND nodes.right < {self.right} AND depth == {self.depth+1} ORDER BY (nodes.right - nodes.left) DESC"
        for record in cursor.execute(sql):
            child_node_id = record[0]
            self.add_child(Node(child_node_id))

    def has_children(self):

        #  root node
        if not self.node_id:
            return True
        return abs(self.left - self.right) > 1


class HpoModel(QAbstractItemModel):
    def __init__(self, conn):
        super().__init__()
        self.conn = conn
        Node.conn = self.conn
        self.root_node = Node(1)

    def columnCount(self, parent=QModelIndex()):
        return 2

    def rowCount(self, parent=QModelIndex()):
        return self.node_from_index(parent).child_count()

    def index(self, row, column, parent: QModelIndex()):

        if not self.hasIndex(row, column, parent):
            return QModelIndex()
        parent_node = self.node_from_index(parent)
        node = parent_node.child(row)

        if node:
            return self.createIndex(row, column, node)
        else:
            return QModelIndex()

    def parent(self, index: QModelIndex()):
        if not index.isValid():
            return QModelIndex()

        node = index.internalPointer()
        parent_node = node.parent

        if parent_node == self.root_node:
            return QModelIndex()

        return self.createIndex(parent_node.row(), 0, parent_node)

    def node_from_index(self, index=QModelIndex()):

        if index == QModelIndex():
            return self.root_node

        return index.internalPointer()

    def data(self, index, role):

        if not index.isValid():
            return None

        node = self.node_from_index(index)

        if role == Qt.DisplayRole:
            if index.column() == 0:
                return node.name

            if index.column() == 1:
                return node.hpo

        return None

    def hasChildren(self, parent: QModelIndex):
        node = self.node_from_index(parent)
        return node.has_children()

    def canFetchMore(self, index: QModelIndex):
        if index == QModelIndex():
            return False

        node = self.node_from_index(index)

        if node.has_children() and node.child_count() == 0:
            return True

        return False

    def fetchMore(self, parent: QModelIndex()):
        node = self.node_from_index(parent)
        count = node.fetch_count()

        self.beginInsertRows(parent, 0, count)
        node.load_child()
        self.endInsertRows()

    def setRoot(self, node: Node):
        self.beginResetModel()

        if self.root_node == None:
            pass  #  delete root node ?

        self.root_node = node
        self.root_node.load_child()
        self.endResetModel()


class HpoQueryWidget(QueryPluginWidget):
    def __init__(self):
        super().__init__()

        self.setWindowTitle("HPO")

        conn = sqlite3.connect("/home/schutz/Dev/hpo2sqlite/hpo.db")
        self.view = QTreeView()
        self.model = HpoModel(conn)
        self.model.setRoot(Node(1))
        self.view.setModel(self.model)

        v_layout = QVBoxLayout()
        v_layout.addWidget(self.view)
        v_layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(v_layout)

    def on_change_query(self):
        pass



# Cutevariant is a light standalone viewer of genetic variation written
# in Python for Qt. It allows to view and filter VCF and other format files.
# Copyright (C) 2018-2019  Labsquare.org
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Please send bugreports with examples or suggestions to
# https://github.com/labsquare/cutevariant/issues

# Standard imports
import sys
import os
from PySide2.QtCore import (
    QCoreApplication,
    QSettings,
    QTranslator,
    QCommandLineParser,
    QCommandLineOption,
)
from PySide2.QtWidgets import QApplication

# Custom imports
from cutevariant.gui import MainWindow, setFontPath, style
import cutevariant.commons as cm
from cutevariant import __version__


def main():
    """The main routine."""

    # Define the names of the organization and the application
    # The value is used by the QSettings class when it is constructed using
    # the empty constructor. This saves having to repeat this information
    # each time a QSettings object is created.
    # The default scope is QSettings::UserScope
    QCoreApplication.setOrganizationName("labsquare")
    QCoreApplication.setApplicationName("cutevariant")
    QCoreApplication.setApplicationVersion(__version__)

    app = QApplication(sys.argv)
    # Process command line arguments
    process_arguments(app)

    # apply dark style 
    style.dark(app)



    # Set icons set
    setFontPath(os.path.join(cm.DIR_FONTS, "materialdesignicons-webfont.ttf"))

    # Translations
    load_translations(app)

    # debug settings
    # from cutevariant.gui.settings import *
    # w = SettingsWidget()
    # w.show()

    # Display
    w = MainWindow()

    STYLES = cm.DIR_STYLES + "frameless.qss"
    with open(STYLES,"r") as file:
        w.setStyleSheet(file.read())

    w.show()
    app.exec_()


def load_translations(app):
    """Load translations

    .. note:: Init ui/locale setting
    """

    # Load locale setting if exists
    # English is the default language
    app_settings = QSettings()
    locale_name = app_settings.value("ui/locale", "en")
    app_translator = QTranslator(app)

    if app_translator.load(locale_name, directory=cm.DIR_TRANSLATIONS):
        app.installTranslator(app_translator)
    else:
        # Init setting
        app_settings.setValue("ui/locale", "en")


def process_arguments(app):
    """Arguments parser"""
    parser = QCommandLineParser()
    # -h, --help, -? (on windows)
    parser.addHelpOption()
    # --version
    show_version = QCommandLineOption(
        ["version"],
        QCoreApplication.translate(
            "main", "Display the version of Cutevariant and exit."
        ),
    )
    parser.addOption(show_version)
    # -v, --verbose
    modify_verbosity = QCommandLineOption(
        ["v", "verbose"],
        QCoreApplication.translate("main", "Modify verbosity."),
        "notset|debug|info|error",  # options available (value name)
        "notset",  # default value
    )
    parser.addOption(modify_verbosity)

    # Process the actual command line arguments given by the user
    parser.process(app)
    # args = parser.positionalArguments()

    if parser.isSet(show_version):
        print("Cutevariant " + __version__)
        exit()

    if parser.isSet(modify_verbosity):
        # Set log level
        cm.log_level(parser.value(modify_verbosity).upper())


if __name__ == "__main__":

    main()



# Standard imports
from logging.handlers import RotatingFileHandler
import logging
import datetime as dt
import tempfile
from pkg_resources import resource_filename

# Misc
MAX_RECENT_PROJECTS = 5
MIN_COMPLETION_LETTERS = 1
DEFAULT_SELECTION_NAME = "variants"

# Paths
DIR_LOGS = tempfile.gettempdir() + "/"

DIR_ASSETS = resource_filename(__name__, "assets/")  # current package name
DIR_TRANSLATIONS = DIR_ASSETS + "i18n/"
DIR_FONTS = DIR_ASSETS + "fonts/"
DIR_ICONS = DIR_ASSETS + "icons/"
DIR_STYLES= DIR_ASSETS + "styles/"

GENOTYPE_ICONS = {
    0: 0xf130,
    1: 0xfaa0,
    2: 0Xfaa4,
    -1: 0xf625
}

# Websites and variant query
WEBSITES_URLS = {"varsome": "https://varsome.com/variant/hg19/{chr}-{pos}-{ref}-{alt}"}


# Logging
LOGGER_NAME = "cutevariant"
LOG_LEVEL = "WARNING"
LOG_LEVELS = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "notset": logging.NOTSET,
}

################################################################################


def logger(name=LOGGER_NAME, logfilename=None):
    """Return logger of given name, without initialize it.

    Equivalent of logging.getLogger() call.
    """
    return logging.getLogger(name)


_logger = logging.getLogger(LOGGER_NAME)
_logger.setLevel(LOG_LEVEL)

# log file
formatter = logging.Formatter("%(asctime)s :: %(levelname)s :: %(message)s")
file_handler = RotatingFileHandler(
    DIR_LOGS
    + LOGGER_NAME
    + "_"
    + dt.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    + ".log",
    "a",
    100_000_000,
    1,
)
file_handler.setLevel(LOG_LEVEL)
file_handler.setFormatter(formatter)
_logger.addHandler(file_handler)

# terminal log
stream_handler = logging.StreamHandler()
formatter = logging.Formatter("%(levelname)s: %(message)s")
stream_handler.setFormatter(formatter)
stream_handler.setLevel(LOG_LEVEL)
_logger.addHandler(stream_handler)


def log_level(level):
    """Set terminal/file log level to given one.
    .. note:: Don't forget the propagation system of messages:
        From logger to handlers. Handlers receive log messages only if
        the main logger doesn't filter them.
    """
    # Main logger
    _logger.setLevel(level.upper())
    # Handlers
    [
        handler.setLevel(level.upper())
        for handler in _logger.handlers
        if handler.__class__
        in (logging.StreamHandler, logging.handlers.RotatingFileHandler)
    ]



from .abstractreader import AbstractReader
import vcf


class FakeReader(AbstractReader):
    def __init__(self):
        super().__init__(None)

    def get_variants(self):
        yield {
            "chr": "11",
            "pos": 125010,
            "ref": "T",
            "alt": "A",
            "annotations": [
                {"transcript": "NM_234234", "gene": "CFTR"},
                {"transcript": "NM_234235", "gene": "CFTR"},
            ],
            "samples": [{"name": "sacha", "gt": 1}],
        }

        yield {
            "chr": "12",
            "pos": 125010,
            "ref": "T",
            "alt": "A",
            "annotations": [
                {"transcript": "NM_234234", "gene": "CFTR"},
                {"transcript": "NM_234235", "gene": "CFTR"},
            ],
            "samples": [{"name": "sacha", "gt": 1}],
        }

        yield {
            "chr": "13",
            "pos": 125010,
            "ref": "T",
            "alt": "A",
            "annotations": [
                {"transcript": "NM_234234", "gene": "CFTR"},
                {"transcript": "NM_234235", "gene": "CFTR"},
            ],
            "samples": [{"name": "sacha", "gt": 1}],
        }

    def get_fields(self):
        """Extract fields informations from VCF fields

        .. note:: Fields used in PRIMARY KEYS have the constraint NOT NULL.
            By default, all other fields can have NULL values.
        """

        yield {
            "name": "chr",
            "category": "variants",
            "description": "chromosom",
            "type": "str",
            "constraint": "NOT NULL",
        }
        yield {
            "name": "pos",
            "category": "variants",
            "description": "position",
            "type": "int",
            "constraint": "NOT NULL",
        }

        yield {
            "name": "ref",
            "category": "variants",
            "description": "reference base",
            "type": "str",
            "constraint": "NOT NULL",
        }
        yield {
            "name": "alt",
            "category": "variants",
            "description": "alternative base",
            "type": "str",
            "constraint": "NOT NULL",
        }

        yield {
            "name": "gt",
            "category": "samples",
            "description": "genotype",
            "type": "int",
        }

        yield {
            "name": "af",
            "category": "samples",
            "description": "allele frequency",
            "type": "float",
        }

        yield {
            "name": "gene",
            "category": "annotations",
            "description": "gene name",
            "type": "str",
        }

        yield {
            "name": "transcript",
            "category": "annotations",
            "description": "gene transcripts",
            "type": "str",
        }

    def get_samples(self):
        return ["sacha"]



# Standard imports
import csv

# Custom imports
from .abstractreader import AbstractReader
from .annotationparser import VEP_ANNOTATION_DEFAULT_FIELDS, BaseParser
from cutevariant.commons import logger

LOGGER = logger()


class CsvReader(AbstractReader):
    """VEP parser to extract data from CSV file

    .. seealso:: AbstractReader class for more information.

    .. note:: About VEP file format:

    http://www.ensembl.org/info/docs/tools/vep/script/vep_other.html#pick

    if a variant overlaps a gene with multiple alternate splicing variants
    (transcripts), then a block of annotation for each of these transcripts
    is reported in the output. In the default VEP output format each of these
    blocks is written on a single line of output; in VCF output format the
    blocks are separated by commas in the INFO field.

    #Uploaded_variation   Location   Allele   Consequence   IMPACT
    SYMBOL   Gene   Feature_type   Feature   BIOTYPE   EXON   INTRON   HGVSc
    HGVSp   cDNA_position   CDS_position   Protein_position
    Amino_acids   Codons   Existing_variation   DISTANCE
    STRAND   FLAGS   SYMBOL_SOURCE   HGNC_ID   TSL   APPRIS
    REFSEQ_MATCHGIVEN_REF   USED_REF   BAM_EDIT
    SIFT   PolyPhen   AF   CLIN_SIG   SOMATIC   PHENO
    PUBMED   MOTIF_NAME   MOTIF_POS   HIGH_INF_POS   MOTIF_SCORE_CHANGE
    LRT_pred   LRT_score   MutationTaster_model   MutationTaster_pred
    SIFT_pred   SIFT_score   clinvar_clnsig   clinvar_rs   clinvar_trait
    """

    def __init__(self, device):
        super().__init__(device)

        # Quick tests on the input file...
        first_line = device.readline()
        csv_dialect = csv.Sniffer().sniff(first_line)

        # Is this file has a correct header ?
        # Test 2 first lines
        header = csv.Sniffer().has_header(first_line + device.readline())
        if not header:
            raise Exception("No header detected in the file; not a CSV file?")

        # Build a csv reader
        self.device.seek(0)
        self.csv_reader = csv.DictReader(self.device, dialect=csv_dialect)

        # hacky hacky...
        # We use BaseParser here only for its function handle_descriptions()
        # that generate full description of annotation fields by taking into
        # account default supported fields set by with VEP_ANNOTATION_DEFAULT_FIELDS
        self.annotation_parser = BaseParser()
        self.annotation_parser.annotation_default_fields = VEP_ANNOTATION_DEFAULT_FIELDS

        # Some columns are ignored or mapped manually to standard ones
        # (chr, pos, ref, alt)
        self.ignored_columns = (
            "location",
            "allele",
            "#uploaded_variation",
            "given_ref",
            "used_ref",
        )

        LOGGER.debug(
            "CsvReader::init: CSV fields found: %s", self.csv_reader.fieldnames
        )

    def __del__(self):
        del self.device

    def get_fields(self):
        """Get full fields descriptions

        .. note:: Annotations fields are added here if they exist in the file.

        .. seealso:: parse_fields()

        :return: Tuple of fields.
            Each field is a dict with the following keys:
                name, category, description, type
            Some fields have an additional constraint key when they are destined
            to be a primary key in the database.
            Annotations fields are added here if they exist in the file.
            .. seealso parse_fields() for basic default fields.
        :rtype: <tuple <dict>>
        """
        LOGGER.debug("CsvReader::get_fields: called")
        if not self.fields:
            LOGGER.debug("CsvReader::get_fields: parse")
            self.fields = tuple(self.parse_fields())
        return self.fields

    def get_variants(self):
        """Get variants as an iterable of dictionnaries

        "annotations" key is added here with the list of annotations if
        they exist in the file.

        .. seealso:: parse_variant()

        :return: Generator of full variants with "annotations" key.
        :rtype: <generator <dict>>
        """
        yield from self.parse_variants()

    def get_samples(self):
        """???????"""
        return []

    def parse_fields(self):
        yield {
            "name": "chr",
            "category": "variants",
            "description": "chromosom",
            "type": "str",
            "constraint": "NOT NULL",
        }
        yield {
            "name": "pos",
            "category": "variants",
            "description": "position",
            "type": "int",
            "constraint": "NOT NULL",
        }
        yield {
            "name": "ref",
            "category": "variants",
            "description": "reference base",
            "type": "str",
            "constraint": "NOT NULL",
        }
        yield {
            "name": "alt",
            "category": "variants",
            "description": "alternative base",
            "type": "str",
            "constraint": "NOT NULL",
        }

        # Get all fields (except the ignored ones)
        raw_fields = (
            field
            for field in self.csv_reader.fieldnames
            if field.lower() not in self.ignored_columns
        )
        # yield full remaped field
        # Hack fix this... the Base class should possess this attribute
        self.annotation_parser.annotation_field_name = list()
        yield from self.annotation_parser.handle_descriptions(raw_fields)

    def parse_variants(self):
        """Read file and parse variants

        .. todo:: An estimation of the progression is made here by updating
            self.read_bytes attribute.

        .. todo: Handle samples

        .. note:: Each line is a transcript; there are many transcripts per variant.

        :return: Generator of variants.
        :rtype: <generator <dict>>
        """

        def add_annotation_to_variant():
            """Add annotation to the current variant"""
            # No annotation parsed
            if not annotation:
                return
            annotations = variant.get("annotations")
            if annotations:
                # Update
                annotations.append(annotation)
            else:
                # Set
                variant["annotations"] = [annotation]

        if not hasattr(self.annotation_parser, "annotation_field_name"):
            raise Exception("Cannot parse variant without parsing first fields")

        variants = dict()
        for i, row in enumerate(self.csv_reader, 1):

            # Build primary key by mapping existing fields in the original file
            chr, pos = self.location_to_chr_pos(row["Location"])
            ref = row["GIVEN_REF"]
            alt = row["Allele"]

            if "USED_REF" in row:
                # Check consistency
                assert row["GIVEN_REF"] == row["USED_REF"], "GIVEN_REF != USED_REF"

            primary_key = (chr, pos, ref, alt)
            # Get previous variant with same primary key (previous transcript)
            variant = variants.get(primary_key, dict())

            # filtre les champs non voulus, cherche ceux qui restent dans
            # VEP_ANNOTATION_DEFAULT_FIELDS pour savoir s'ils sont supportés;
            # sinon les ajoute en lower case en fallback
            annotation = dict()
            # Remove unwanted fields
            g = (key for key in row.keys() if key.lower() not in self.ignored_columns)
            for raw_key in g:
                lower_key = raw_key.lower()
                # Get supported fields
                field_descript = VEP_ANNOTATION_DEFAULT_FIELDS.get(lower_key)
                if field_descript:
                    # Use supported field name
                    lower_key = field_descript["name"]

                annotation[lower_key] = row[raw_key]

            # Quicker ?
            # annotation = {
            #    VEP_ANNOTATION_DEFAULT_FIELDS.get(key.lower(), {'name': key.lower()})['name']: value
            #    for key, value in row.items() if key.lower() not in self.ignored_columns
            # }

            if variant:
                # Variant already created
                # Append current annotation to the previous ones
                add_annotation_to_variant()
                continue

            # New variant
            variant["chr"], variant["pos"], variant["ref"], variant["alt"] = primary_key

            # Set current annotation
            add_annotation_to_variant()

            # testing purpose
            #            variant["samples"] = [
            #                {"name": "boby", "gt": 0},
            #                {"name": "sacha", "gt": 1},
            #                {"name": "olivier", "gt": 2},
            #            ]

            variants[primary_key] = variant

        LOGGER.info(
            "CsvReader::parse_variants: transcripts %s, variants %s", i, len(variants)
        )

        for variant in variants.values():
            yield dict(variant)

    def location_to_chr_pos(self, location: str):
        """Parse VEP `Location` field to extract `chr`, `pos` fields

        2 representations can be encountered in standard coordinate format
            - chr:start
            - chr:start-end

        The start position is always returned as `pos` value.

        :Example:
            11:10000-10000 => chr, pos => 11, 10000

        :param location: Location field in a VEP file.
        :return: Tuple of chr and pos data.
        :rtype: <tuple <str>, <str>>
        """
        chr, positions = location.split(":")
        pos = positions.split("-")[0]
        return chr, pos

    def __repr__(self):
        return f"VEP Reader using {type(self.annotation_parser).__name__}"



"""Plugin to view/edit/remove/do set operations on selections in the database
from the GUI.

SelectionQueryWidget class is seen by the user and uses SelectionQueryModel class
as a model that handles records from the database.
"""
# Qt imports
from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *

# Custom imports
from .plugin import QueryPluginWidget
from cutevariant.core import Query
from cutevariant.core import sql
from cutevariant.gui.ficon import FIcon
from cutevariant.commons import logger, DEFAULT_SELECTION_NAME


LOGGER = logger()

# =================== SELECTION MODEL ===========================
class SelectionQueryModel(QAbstractTableModel):
    """Model to store all selections from SQLite `selections` table.

    Usage:

        model = SelectionQueryModel()
        model.query = query
        model.load()

    """

    def __init__(self):
        super().__init__()
        self._query = None
        self.records = []

    def rowCount(self, parent=QModelIndex()):
        """Overrided from QAbstractTableModel"""
        return len(self.records)

    def columnCount(self, parent=QModelIndex()):
        """Overrided from QAbstractTableModel"""
        return 2  #  value and count

    def data(self, index: QModelIndex(), role=Qt.DisplayRole):
        """Return the data stored under the given role for the item referred
        to by the index (row, col)

        Overrided from QAbstractTableModel

        :param index:
        :param role:
        :type index:
        :type role:
        :return: None if no valid index
        :rtype:
        """

        if not index.isValid():
            return None

        if role == Qt.DisplayRole:
            if index.column() == 0:
                return self.records[index.row()]["name"]

            if index.column() == 1:
                return self.records[index.row()]["count"]

        return None

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Return the data for the given role and section in the header with
        the specified orientation

        Overrided from QAbstractTableModel
        """
        if not self.records:
            return

        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if section == 0:
                return "selection"
            if section == 1:
                return "count"

        if orientation == Qt.Vertical and role == Qt.DisplayRole and section > 0:
            return self.records[section].get(
                "id", None
            )  # For debug purpose . displayed in vertical header

    def record(self, index: QModelIndex()):
        """Return Selection records by index"""
        if not index.isValid():
            return None
        return self.records[index.row()]

    def find_record(self, name: str):
        """Find a record by name
        @see: view.selectionModel()
        :return: Tuple of index in the model, and the record itself.
        :rtype: <tuple <int>, <str>>
        """
        for idx, record in enumerate(self.records):
            if record["name"] == name:
                return idx, record
        return None

    def remove_record(self, index: QModelIndex()):
        """Delete the selection with the given id in the database

        :return: True if the deletion has been made, False otherwise.
        :rtype: <boolean>
        """
        # Get selected record
        record = self.record(index)
        # Delete in database
        if sql.delete_selection(self.query.conn, record["id"]):
            # Delete in model; triggers currentRowChanged signal
            self.beginRemoveRows(QModelIndex(), index.row(), index.row())
            # Delete in records
            self.records.pop(index.row())
            self.endRemoveRows()
            return True
        return False

    def edit_record(self, index, record: dict):
        """Edit the given selection in the database and emit `dataChanged` signal"""
        if sql.edit_selection(self.query.conn, record):
            self.dataChanged.emit(index, index)

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, query: Query):
        self._query = query

    def load(self):
        """Load all selections into the model"""
        self.beginResetModel()
        # Add all selections from the database
        # Dictionnary of all attributes of the table.
        #    :Example: {"name": ..., "count": ..., "query": ...}
        self.records = list(sql.get_selections(self._query.conn))
        self.endResetModel()

    def save_current_query(self, name):
        """Save current query as a new selection and reload the model"""
        # TODO: just get the id of the created selection and add a new record
        # instead of reloading all the model...
        if self.query.create_selection(name):
            self.load()


# =================== SELECTION VIEW ===========================


class SelectionQueryWidget(QueryPluginWidget):
    """Widget displaying the list of avaible selections.
    User can select one of them to update Query::selection
    """

    def __init__(self):
        super().__init__()

        self.setWindowTitle(self.tr("Selections"))

        self.model = SelectionQueryModel()
        self.view = QTableView()
        self.view.setModel(self.model)
        self.view.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.view.setSelectionMode(QAbstractItemView.SingleSelection)
        self.view.horizontalHeader().hide()
        # self.view.horizontalHeader().setStretchLastSection(True)

        self.view.verticalHeader().show()
        self.view.verticalHeader().setDefaultSectionSize(26)
        self.view.setShowGrid(False)
        self.view.setAlternatingRowColors(True)

        layout = QVBoxLayout()
        layout.addWidget(self.view)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

        # call on_current_row_changed when item selection changed
        self.view.selectionModel().currentRowChanged.connect(
            self.on_current_row_changed
        )

    def menu_setup(self, locked_selection=False):
        """Setup popup menu
        :key locked_selection: Allow to mask edit/remove actions (default False)
        :type locked_selection: <boolean>
        """
        menu = QMenu()

        if not locked_selection:
            menu.addAction(FIcon(0xF8FF), self.tr("Edit"), self.edit_selection)

        # Set operations on selections: create mapping and actions
        set_icons_ids = (0xF55D, 0xF55B, 0xF564)
        set_texts = (self.tr("Intersect"), self.tr("Difference"), self.tr("Union"))
        set_internal_ids = ("intersect", "difference", "union")
        # Map the operations with an internal id not visible for the user
        # This id is used by _create_set_operation_menu and _make_set_operation
        # Keys: user text; values: internal ids
        self.set_operations_mapping = dict(zip(set_texts, set_internal_ids))

        # Create actions
        [
            menu.addMenu(self._create_set_operation_menu(FIcon(icon_id), text))
            for icon_id, text in zip(set_icons_ids, set_texts)
        ]

        if not locked_selection:
            menu.addSeparator()
            menu.addAction(FIcon(0xF413), self.tr("Remove"), self.remove_selection)
        return menu

    def load(self):
        """Load selection model and update the view"""
        # Block signals during the insertions
        self.view.selectionModel().blockSignals(True)
        self.model.load()
        self.view.horizontalHeader().setSectionResizeMode(0, QHeaderView.Stretch)
        self.view.horizontalHeader().setSectionResizeMode(
            1, QHeaderView.ResizeToContents
        )

        # Select record according to query.selection
        row, _ = self.model.find_record(self.query.selection)
        if row is not None:
            self.view.selectRow(row)

        self.view.selectionModel().blockSignals(False)

    def on_change_query(self):
        """Overrided from AbstractQueryWidget"""
        self.load()

    def on_init_query(self):
        """Overrided from AbstractQueryWidget"""
        self.model.query = self.query

    def on_current_row_changed(self, index):
        """Update query when a selection item is clicked

        .. note:: Slot called when item selection is changed.
        """
        # We don't really care of the query that created the selection
        # The joins are based on the name of the table
        self.query.selection = self.model.record(index)["name"]
        self.query_changed.emit()

    def save_current_query(self):
        """Open a dialog box to save the current query into a selection"""
        name, success = QInputDialog.getText(
            self,
            self.tr("Type a name for selection"),
            self.tr("Selection name:"),
            QLineEdit.Normal,
        )
        if not success:
            return

        if name == DEFAULT_SELECTION_NAME:
            LOGGER.error(
                "SelectionQueryWidget:save_current_query:: '%s' is a reserved name for a selection.",
                name,
            )
            self.message.emit(
                self.tr("'%s' is a reserved name for a selection!") % name
            )
        elif name in {record["name"] for record in self.model.records}:
            LOGGER.error(
                "SelectionQueryWidget:save_current_query:: '%s' is a already used.",
                name,
            )
            self.message.emit(self.tr("'%s' is a already used for a selection!") % name)
        else:
            self.model.save_current_query(name)

    def contextMenuEvent(self, event: QContextMenuEvent):
        """Overrided: Show popup menu at the cursor position"""
        if not self.model.records:
            return
        # Detect locked selection to mask edit/remove actions
        locked_selection = False
        current_index = self.view.selectionModel().currentIndex()
        record = self.model.record(current_index)
        if record["name"] == DEFAULT_SELECTION_NAME:
            locked_selection = True
        # Show the menu
        menu = self.menu_setup(locked_selection)
        menu.exec_(event.globalPos())

    def _create_set_operation_menu(self, icon, menu_name):
        """Dynamically add submenu with the given name to popup menu"""
        menu = QMenu(menu_name)
        menu.setIcon(icon)

        # Get all the names of selections except the current one
        current_selection_name = self.model.record(self.view.currentIndex())["name"]
        records_names = set(record["name"] for record in self.model.records)
        records_names.remove(current_selection_name)

        # Create an action in the menu for all the remaining elements
        for record_name in records_names:
            action = menu.addAction(record_name)
            # This hidden data is used by _make_set_operation() to detect
            # which operation to do
            action.setData(self.set_operations_mapping[menu_name])
            action.triggered.connect(self._make_set_operation)

        return menu

    def _make_set_operation(self):
        """Do set operation

        .. note:: Called when a "set" submenu of the popup menu is triggered.
        """
        action = self.sender()

        # Get menu id to know which set operation to do
        set_internal_id = action.data()

        # Init class attribute
        sql.Selection.conn = self.query.conn

        # Get the records and extract their database id to build 2 Selections objects
        record_1 = self.model.record(self.view.selectionModel().currentIndex())
        _, record_2 = self.model.find_record(action.text())
        # TODO: handle modes for creation of selections
        selection_1 = sql.Selection.from_selection_id(record_1["id"])
        selection_2 = sql.Selection.from_selection_id(record_2["id"])

        new_selection_name, success = QInputDialog.getText(
            self, self.tr("Type a name for selection"), self.tr("Selection name:")
        )
        if not success:
            return

        # Do set operation and create new selection
        if set_internal_id == "union":
            selection_3 = selection_1 + selection_2
        if set_internal_id == "difference":
            selection_3 = selection_1 - selection_2
        if set_internal_id == "intersect":
            selection_3 = selection_1 & selection_2

        LOGGER.debug(
            "Query:_make_set_operation:: New selection query: %s", selection_3.sql_query
        )

        if not selection_3.save(new_selection_name):
            self.message.emit(self.tr("Fail to create the selection!"))
            return
        # Reload the model
        self.model.load()

    def remove_selection(self):
        """Remove a selection from the database"""
        msg = QMessageBox()
        msg.setText(self.tr("Are you sure you want to remove this selection?"))
        msg.setStandardButtons(QMessageBox.Yes | QMessageBox.No)

        if msg.exec_() == QMessageBox.Yes:
            for index in self.view.selectionModel().selectedRows():
                self.model.remove_record(index)

    def edit_selection(self):
        """Update a selection in the database
        .. note:: We do not reload all the UI for this
        """
        current_index = self.view.selectionModel().currentIndex()
        new_name = QInputDialog.getText(
            self, self.tr("Type a new name"), self.tr("Selection name:")
        )
        if new_name[1] and current_index:
            old_record = self.model.record(current_index)
            old_record["name"] = new_name[0]
            self.model.edit_record(current_index, old_record)



# Standard imports
import os
from PySide2.QtWidgets import *
from PySide2.QtCore import QThread, Signal, QDir, QSettings, QFile
from PySide2.QtGui import QIcon

# Custom imports
from cutevariant.core.importer import async_import_file
from cutevariant.core import get_sql_connexion
import cutevariant.commons as cm
from cutevariant.core.readerfactory import detect_vcf_annotation

LOGGER = cm.logger()


class ProjectPage(QWizardPage):
    """Page: Creation of a new project
    We ask the reference genome, the name and the path of the future project file.
    """

    def __init__(self):
        super().__init__()

        self.setTitle(self.tr("Project creation"))
        self.setSubTitle(
            self.tr("This wizard will guide you to create a cutevariant project.")
        )

        self.project_name_edit = QLineEdit()
        self.project_path_edit = QLineEdit()
        self.project_path_edit.setText(os.getcwd())
        self.browse_button = QPushButton("Browse")
        self.reference = QComboBox()

        self.reference.addItem("hg19")
        self.registerField("project_name", self.project_name_edit, "text")
        self.registerField("project_path", self.project_path_edit, "text")
        self.registerField("reference", self.reference, "currentText")

        v_layout = QFormLayout()

        browse_layout = QHBoxLayout()
        browse_layout.addWidget(self.project_path_edit)
        browse_layout.addWidget(self.browse_button)
        browse_layout.setContentsMargins(0, 0, 0, 0)

        v_layout.addRow(self.tr("Reference genom"), self.reference)
        v_layout.addRow(self.tr("Project Name"), self.project_name_edit)
        v_layout.addRow(self.tr("Create in"), browse_layout)

        self.setLayout(v_layout)

        self.browse_button.clicked.connect(self._browse)
        self.project_path_edit.textChanged.connect(self.completeChanged)
        self.project_name_edit.textChanged.connect(self.completeChanged)

    def _browse(self):
        """Open a dialog box to set the directory where the project will be saved"""
        path = QFileDialog.getExistingDirectory(
            self, self.tr("Select a path for the project")
        )
        if path:
            self.project_path_edit.setText(path)

    def isComplete(self):
        """Conditions to unlock next button"""
        return (
            True
            if (
                QDir(self.project_path_edit.text()).exists()
                and self.project_path_edit.text()
                and self.project_name_edit.text()
            )
            else False
        )


class FilePage(QWizardPage):
    """Page: Open the file containing variant data
    We ask the path of the data file
    """

    def __init__(self):
        super().__init__()

        self.setTitle(self.tr("Select a file"))
        self.setSubTitle(self.tr("Supported file are vcf, vcf.gz, vep.txt."))

        self.file_path_edit = QLineEdit()
        self.anotation_detect_label = QLabel()
        self.button = QPushButton(self.tr("Browse"))
        h_layout = QHBoxLayout()

        h_layout.addWidget(self.file_path_edit)
        h_layout.addWidget(self.button)

        self.v_layout = QVBoxLayout()
        self.v_layout.addLayout(h_layout)
        self.v_layout.addWidget(self.anotation_detect_label)
        self.setLayout(self.v_layout)

        self.button.clicked.connect(self._browse)
        self.file_path_edit.textChanged.connect(self.completeChanged)

        self.registerField("filename", self.file_path_edit, "text")
        #  annotation ? should be an option or not ?
        self.registerField("annotation", self.anotation_detect_label, "text")

    def _browse(self):
        """Open a dialog box to set the data file

        The file is opened and we show an alert if annotations are detected.
        """
        # Reload last directory used
        app_settings = QSettings()
        last_directory = app_settings.value("last_directory", QDir.homePath())

        filepath, filetype = QFileDialog.getOpenFileName(
            self,
            self.tr("Open a file"),
            last_directory,
            self.tr("VCF file (*.vcf *.vcf.gz);; CSV file (*.csv *.tsv *.txt)"),
        )

        if filepath:
            # Display and save directory
            self.file_path_edit.setText(filepath)
            app_settings.setValue("last_directory", os.path.dirname(filepath))

            if "vcf" in filepath:
                # TODO: detect annotations on other tyes of files...
                annotation_type = detect_vcf_annotation(filepath)
                if annotation_type:
                    text = self.tr("<b>%s annotations detected!</b>") % annotation_type
                else:
                    text = self.tr("<b>No annotation data has been detected!</b>")

                self.anotation_detect_label.setText(text)

    def isComplete(self):
        """Conditions to unlock next button"""
        return (
            True
            if (
                self.file_path_edit.text()
                and QFile(self.file_path_edit.text()).exists()
            )
            else False
        )


class ImportThread(QThread):
    """Thread used to create a new project by importing a variant file

    .. note:: This thread creates the database.
    .. seealso:: ImportPage class.
    """

    # Qt signals
    progress_changed = Signal(int, str)
    finished_status = Signal(bool)

    def __init__(self):
        super().__init__()
        self._stop = False
        # File top open
        self.filename = ""
        # Project's filepath
        self.db_filename = ""
        self.project_settings = dict()

    def set_importer_settings(self, filename, db_filename, project_settings={}):
        """Init settings of the importer

        :param filename: File to be opened.
        :param db_filename: Filepath of the new project.
        :key project_settings: The reference genome and the name of the project.
            Keys have to be at least "reference" and "project_name".
        :type filename: <str>
        :type db_filename: <str>
        :type project_settings: <dict>
        """
        # File top open
        self.filename = filename
        # Project's filepath
        self.db_filename = db_filename
        # Project settings
        self.project_settings = project_settings

    def run(self):
        """Overrided QThread method

        .. warning:: This method is the only one to be executed in the thread.
            SQLite objects created in a thread can only be used in that same
            thread; so no use of self.conn is allowed elsewhere.
        """
        self._stop = False

        if os.path.exists(self.db_filename):
            os.remove(self.db_filename)
        self.conn = get_sql_connexion(self.db_filename)

        try:
            # Import the file
            for value, message in async_import_file(
                self.conn, self.filename, self.project_settings
            ):
                if self._stop == True:
                    self.conn.close()
                    break
                # Send progression
                self.progress_changed.emit(value, message)
        except BaseException as e:
            self.progress_changed.emit(0, str(e))
            self._stop = True
            raise e
        finally:
            # Send status (Send True when there is no error)
            self.finished_status.emit(not self._stop)

    def stop(self):
        self._stop = True


class ImportPage(QWizardPage):
    """Page: Creation of the database"""

    def __init__(self):
        super().__init__()

        self.setTitle(self.tr("Import file"))
        self.setSubTitle(
            self.tr("Please click on Import/Stop to start or stop the process.")
        )
        self.text_buttons = [self.tr("Import"), self.tr("Stop")]

        self.thread_finished = False
        self.thread = ImportThread()
        self.progress = QProgressBar()
        self.import_button = QPushButton(self.text_buttons[0])

        v_layout = QHBoxLayout()
        v_layout.addWidget(self.progress)
        v_layout.addWidget(self.import_button)

        m_layout = QVBoxLayout()

        self.log_edit = QPlainTextEdit()
        self.tab_widget = QTabWidget()
        self.tab_widget.addTab(self.log_edit, "log")

        m_layout.addLayout(v_layout)
        m_layout.addWidget(self.tab_widget)
        self.setLayout(m_layout)

        # File to open
        self.log_edit.appendPlainText(self.field("filename"))

        self.thread.started.connect(
            lambda: self.log_edit.appendPlainText(self.tr("Started"))
        )
        self.import_button.clicked.connect(self.run)
        self.thread.progress_changed.connect(self.progress_changed)
        self.thread.finished.connect(self.import_thread_finished)
        self.thread.finished_status.connect(self.import_thread_finished_status)

    def progress_changed(self, value, message):
        """Update the progress bar
        Slot called when progress_changed is emitted by the thread
        """
        self.progress.setValue(value)
        if message:
            self.log_edit.appendPlainText(message)

    def import_thread_finished(self):
        """Force the activation of the finish button after a successful import"""
        self.completeChanged.emit()

    def import_thread_finished_status(self, status):
        """Set the finished status of import thread

        .. note:: Called at the end of run()
        """
        self.thread_finished = status
        if status:
            # Block further import
            self.import_button.setDisabled(True)
            self.log_edit.appendPlainText(self.tr("Done"))
        else:
            # Display import on the button
            self.import_button.setText(self.text_buttons[0])
            self.log_edit.appendPlainText(self.tr("Stopped!"))

    def run(self):
        """Called when import button is clicked
        Launch the import in a separate thread
        """
        if self.thread.isRunning():
            LOGGER.debug("ImportPage:run: stop thread")
            self.import_button.setDisabled(True)
            self.thread.stop()
            self.progress.setValue(0)
            self.import_button.setDisabled(False)
            self.import_button.setText(self.text_buttons[0])

        else:
            self.thread.set_importer_settings(
                # File to open
                self.field("filename"),
                # Project's filepath
                (
                    self.field("project_path")
                    + QDir.separator()
                    + self.field("project_name")
                    + ".db"
                ),
                # Project's settings
                {
                    # Reference genome
                    "reference": self.field("reference"),
                    # Project's name
                    "project_name": self.field("project_name"),
                },
            )

            self.log_edit.appendPlainText(self.tr("Import ") + self.thread.filename)
            # display stop on the button
            self.import_button.setText(self.text_buttons[1])
            self.thread.start()

    def isComplete(self):
        """Conditions to unlock finish button"""
        return self.thread_finished


class ProjectWizard(QWizard):
    """Main window of the project wizard

    3 pages are instantiated here:
        - ProjectPage: Creation of a new project
        - FilePage: Open the file containing variant data
        - ImportPage: Creation of the database
    """

    def __init__(self):
        super().__init__()
        self.setWindowTitle(self.tr("Cutevariant - Project creation wizard"))
        self.setWindowIcon(QIcon(cm.DIR_ICONS + "app.png"))

        self.addPage(ProjectPage())
        self.addPage(FilePage())
        self.addPage(ImportPage())
